{
 "cells": [
  {
   "cell_type": "markdown",
   "id": "32bf2e6d",
   "metadata": {},
   "source": [
    "# Week 8 - Metropolis Algorithm and Autocorrelation function\n"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Recap: Metropolis Algorithm\n",
    "\n",
    "Using the **Metropolis algorithm** to sample the distribution\n",
    "\n",
    "$$P(x) = e^{\\frac{-x^2}{2\\sigma^2}}.$$\n",
    "\n",
    "In order to do so, we will create a Python class that runs a Metropolis algorithm reproducing the distribution $P(x)$. We will employ a **Markov chain** obtained through **one random walker** behaving according to the following transition probability:\n",
    "\n",
    "$$T(x_i \\rightarrow x_j) = \\min{\\left[1,P(x_j)/P(x_i)\\right]},$$\n",
    "\n",
    "where the trial move is obtained with a uniformly distributed random variable $\\tilde{x} \\in [0,1) $\n",
    "\n",
    "$$ x_j = x_i + \\delta(\\tilde{x}-0.5) $$\n",
    "\n",
    "which is accepted if \n",
    "\n",
    "$$\\tilde{x}_w \\leq w = P(x_j)/P(x_i) = e^{\\frac{-(x_j^2-x_i^2)}{2\\sigma^2}} $$\n",
    "\n",
    "with $\\tilde{x}_w \\in [0,1)$ also being a random variable following a uniform distribution."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "import numpy as np\n",
    "import matplotlib.pyplot as plt\n",
    "\n",
    "class Metropolis:\n",
    "    '''\n",
    "    This class implements \n",
    "    the Metropolis algorithm for a Gaussian distribution\n",
    "    exp(-x**2/(2*sigma**2))\n",
    "    '''\n",
    "    \n",
    "    def __init__(self,N,delta,sigma=1):\n",
    "        'Initialise the Metropolis'\n",
    "        self.N = N\n",
    "        self.rng = np.random.default_rng(seed=42424)\n",
    "        self.sigma = sigma\n",
    "        self.delta = delta\n",
    "        self.pref = -1.0/(2.0*self.sigma**2)\n",
    "        self.acc = 0\n",
    "        self.x = 0.0\n",
    "        self.ar = None\n",
    "        self.all_x = [self.x]\n",
    "        self.all_x2 = None\n",
    "        self.all_x3 = None\n",
    "        self.all_x4 = None\n",
    "        self.var = None\n",
    "        self.counter = 0\n",
    "        self.autocorr = []\n",
    "        self.n_autocorr = 10\n",
    "    \n",
    "    def metro_step(self):\n",
    "        'One step of Metropolis'\n",
    "        rnd = self.rng.random()\n",
    "        xp = self.x + self.delta * (rnd-0.5)\n",
    "        p = self.x**2\n",
    "        pp = xp**2\n",
    "        w = np.exp(self.pref*(pp-p))\n",
    "        rnd_w = self.rng.random()\n",
    "        if (rnd_w<=w):\n",
    "            self.x = xp \n",
    "            self.acc += 1\n",
    "        self.all_x.append(self.x)\n",
    "    \n",
    "    def metro_driver(self):\n",
    "        'Run N Metropolis steps.'\n",
    "        for i in range(self.N):\n",
    "            self.counter += 1\n",
    "            self.metro_step()\n",
    "    \n",
    "    def get_hist(self):\n",
    "        'Plot Histogram'\n",
    "        plt.xlabel('x')\n",
    "        plt.ylabel('count')\n",
    "        plt.hist(self.all_x,bins=100)\n",
    "        plt.show()\n",
    "    \n",
    "    @staticmethod\n",
    "    def calc_autocorr(a,N):\n",
    "        'Autocorrelation function'\n",
    "        res = np.ones(N)\n",
    "        for k in range(1,N):\n",
    "            c_k = np.average(a[:-k]*a[k:]) \n",
    "            x_r = np.average(a[:-k])\n",
    "            x_r_2 = np.average(a[:-k]**2)\n",
    "            res[k] = (c_k-x_r**2)/(x_r_2-x_r**2)\n",
    "        return res\n",
    "    \n",
    "    def get_stats(self):\n",
    "        '''\n",
    "        Calculate, store and print \n",
    "        statistical quantities\n",
    "        '''\n",
    "        self.ar = self.acc*1.0/self.counter\n",
    "        self.aver_x = np.average(self.all_x)\n",
    "        self.aver_x2 = np.average(np.array(self.all_x)**2)\n",
    "        self.aver_x3 = np.average(np.array(self.all_x)**3)\n",
    "        self.aver_x4 = np.average(np.array(self.all_x)**4)\n",
    "        self.var = self.aver_x2-self.aver_x**2\n",
    "        print(\"Total number of steps run = {}\".format(self.counter))\n",
    "        print(\"Acceptance ratio = {}\".format(self.ar))\n",
    "        print(\"Sigma = {}, delta = {}\".format(self.sigma,self.delta))\n",
    "        print(\"<x> = {}, <x2> = {}\".format(self.aver_x,self.aver_x2))\n",
    "        print(\"<x3> = {}, <x4> = {}\".format(self.aver_x3,self.aver_x4))\n",
    "        print(\"variance = {}\".format(self.var))\n",
    "        \n",
    "    def get_autocorr(self,n_autocorr=100):\n",
    "        '''\n",
    "        Wrapper for the autocorrelation\n",
    "        function calculator\n",
    "        '''\n",
    "        a = np.array(self.all_x)\n",
    "        self.autocorr = self.calc_autocorr(a,n_autocorr)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Autocorrelation function\n",
    "\n",
    "Now let us investigate the autocorrelation function $c(k)$ for our Metropolis algorithm, which is defined as \n",
    "\n",
    "$$c(k) =\\frac{<x_ix_{i+k}>-<x_i>^2}{<x_i^2>-<x_i>^2}.$$"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "e9a77b49",
   "metadata": {},
   "outputs": [],
   "source": [
    "metro = Metropolis(10000,5) \n",
    "metro.metro_driver()\n",
    "metro.get_stats()\n",
    "metro.get_autocorr(n_autocorr=20) #try to increase it\n",
    "plt.xlabel('k')\n",
    "plt.ylabel('c(k)')\n",
    "plt.plot(range(len(metro.autocorr)),metro.autocorr,'o-')\n",
    "plt.axhline(0.0,c='grey',alpha=0.5)\n",
    "plt.show()"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "* Let's us look at the trajectory across different time scales."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "metro.all_x\n",
    "plt.xlim((0,10000)) #<-- Change time scale\n",
    "plt.plot(metro.all_x,'o-',lw=0.5,ms=0.5)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "3bb6dbae",
   "metadata": {},
   "source": [
    "- Now we study the autocorrelation as function of the integration step $\\delta$."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "b6d2742f",
   "metadata": {},
   "outputs": [],
   "source": [
    "for delta in [1,5,10,25,50]:\n",
    "    metro = Metropolis(10000,delta) \n",
    "    metro.metro_driver()\n",
    "    metro.get_stats()\n",
    "    metro.get_autocorr(n_autocorr=20)\n",
    "    plt.xlabel('k')\n",
    "    plt.ylabel('c(k)')\n",
    "    xx = range(len(metro.autocorr))\n",
    "    plt.plot(xx,metro.autocorr,'o-',label='$\\delta =${}'.format(delta))\n",
    "#plt.xlim([0,len(xx)])\n",
    "plt.legend()\n",
    "plt.show()"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "b3d30300",
   "metadata": {},
   "source": [
    "- Now let's try to calculate the same autocorrelation function for a normal distribution by using ad-hoc libraries for random number generators. Is it different with respect to our Metropolis algorithm?"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "1bd25854",
   "metadata": {},
   "outputs": [],
   "source": [
    "rng = np.random.default_rng(seed=42424)\n",
    "N= 1000\n",
    "a = rng.normal(size=N)\n",
    "Nc = 20\n",
    "c = Metropolis.calc_autocorr(a,Nc)\n",
    "plt.xlabel('k')\n",
    "plt.ylabel('c(k)')\n",
    "plt.plot(range(Nc),c,'o-')\n",
    "plt.axhline(0.0,c='grey',alpha=0.5)\n",
    "plt.show()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "05ab33c4",
   "metadata": {},
   "outputs": [],
   "source": [
    "plt.hist(a)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "e39ae15c",
   "metadata": {},
   "outputs": [],
   "source": []
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3 (ipykernel)",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.11.1"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 5
}
